@@ -1,6 +1,6 @@
 ;;; clojure-mode-sexp-test.el --- Clojure Mode: sexp tests  -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2015-2016 Artur Malabarba <artur@endlessparentheses.com>
+;; Copyright (C) 2015-2018 Artur Malabarba <artur@endlessparentheses.com>
 
 ;; This file is not part of GNU Emacs.
 
@@ -75,6 +75,26 @@
     (insert "(+ 10")
     (newline-and-indent)))
 
+(ert-deftest clojure-find-ns-test ()
+  (with-temp-buffer
+    (insert "(ns ^{:doc \"Some docs\"}\nfoo-bar)")
+    (newline)
+    (newline)
+    (insert "(in-ns 'baz-quux)")
+    (clojure-mode)
+
+    ;; From inside docstring of first ns
+    (goto-char 18)
+    (should (equal "foo-bar" (clojure-find-ns)))
+
+    ;; From inside first ns's name, on its own line
+    (goto-char 29)
+    (should (equal "foo-bar" (clojure-find-ns)))
+
+    ;; From inside second ns's name
+    (goto-char 42)
+    (should (equal "baz-quux" (clojure-find-ns)))))
+
 (provide 'clojure-mode-sexp-test)
 
 ;;; clojure-mode-sexp-test.el ends here
